#!/usr/bin/env python
# ******************************************************************************
# Copyright 2019 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
A set of functions to convert a Keras (tf.keras) model to a new
equivalent model with different characteristics. Then, the new model
can be quantized.

"""
import numpy as np
from tensorflow.keras.layers import MaxPooling2D, AveragePooling2D, \
    BatchNormalization, GlobalAveragePooling2D, InputLayer
from tensorflow.keras.models import Model
from .quantization_layers import QuantizedConv2D, QuantizedDense, \
    QuantizedDepthwiseConv2D, QuantizedSeparableConv2D

def invert_batchnorm_pooling(model):
    """Returns a new model where pooling and batchnorm layers are inverted.

    From a Keras model where pooling layers precede batch normalization
    layers, this function places the BN layers before pooling layers. This
    is the first step before folding BN layers into neural layers.

    Note: inversion of layers is equivalent only if the gammas of BN layers
    are positive. The function raises an error if not.

    Args:
        model (:obj:`tf.keras.Model`): a tf.keras model.

    Returns:
        :obj:`tf.keras.Model`: a keras.Model.

    """
    x = model.layers[0].output
    i = 1
    while i < len(model.layers) - 1:
        layer = model.layers[i]
        next_layer = model.layers[i+1]
        supported_pool = (MaxPooling2D,
                          AveragePooling2D,
                          GlobalAveragePooling2D)
        is_pool = isinstance(layer, supported_pool)
        if is_pool and isinstance(next_layer, BatchNormalization):
            gammas = next_layer.get_weights()[0]
            if np.sum(gammas <= 0) > 0:
                # NB: negative gammas are only a problem for max pooling, not
                # for avg pooling
                raise RuntimeError(f"There are {np.sum(gammas <= 0)} negative "
                                   "gammas in the batch norm layer "
                                   "{next_layer.name}. Negative gammas are "
                                   "not supported.")
            next_layer_config = next_layer.get_config()
            # GlobalAveragePooling2D brings a change on axis for the batch norm.
            if isinstance(layer, GlobalAveragePooling2D):
                next_layer_config['axis'] = [-1]
            bn_layer = BatchNormalization.from_config(next_layer_config)
            x = bn_layer(x)
            x = layer(x)
            bn_layer.set_weights(next_layer.get_weights())
            i = i + 2
        else:
            x = layer(x)
            i = i + 1

    x = model.layers[-1](x)
    return Model(inputs=model.input, outputs=[x], name='model_BN_MP')


def fold_batch_norms(model):
    """Returns a new model where batchnorm layers are folded into
    previous neural layers.

    From a Keras model where BN layers follow neural layers, this
    function removes the BN layers and updates weights and bias
    accordingly of the preceding neural layers. The new model is
    strictly equivalent to the previous one.

    Args:
        model (:obj:`tf.keras.Model`): a Keras model.

    Returns:
        :obj:`tf.keras.Model`: a tf.keras.Model.

    """
    x = model.layers[0].output
    i = 1
    while i < len(model.layers) - 1:
        layer = model.layers[i]
        next_layer = model.layers[i+1]

        if isinstance(next_layer, BatchNormalization):
            if not isinstance(layer, (QuantizedConv2D,
                                      QuantizedDepthwiseConv2D,
                                      QuantizedDense)):
                raise AttributeError("The layer preceding a batch norm "
                                     "layer must be QuantizedDense, "
                                     "QuantizedConv2D or "
                                     "QuantizedDepthwiseConv2D.")

            # Get weights and BN parameters
            gamma, beta, mean, var = next_layer.get_weights()
            epsilon = next_layer.epsilon
            weights = layer.get_weights()[0]
            bias = layer.get_weights()[1] if len(
                layer.get_weights()) > 1 else 0

            # Compute new weights for folded layer
            scale_BN = gamma / np.sqrt(var + epsilon)
            new_weights = [weights * scale_BN]
            new_weights.append(beta + (bias-mean) * scale_BN)

            # Create new layer

            if isinstance(layer, QuantizedDepthwiseConv2D):
                new_layer = QuantizedDepthwiseConv2D(layer.kernel_size,
                                                     quantizer=layer.quantizer,
                                                     use_bias=True,
                                                     padding=layer.padding,
                                                     name=layer.name + '_foldBN')
            elif isinstance(layer, QuantizedConv2D):
                new_layer = QuantizedConv2D(layer.filters, layer.kernel_size,
                                            quantizer=layer.quantizer,
                                            strides=layer.strides,
                                            use_bias=True,
                                            padding=layer.padding,
                                            name=layer.name + '_foldBN')
            elif isinstance(layer, QuantizedDense):
                new_layer = QuantizedDense(layer.units,
                                           quantizer=layer.quantizer,
                                           use_bias=True,
                                           name=layer.name + '_foldBN')

            x = new_layer(x)
            new_layer.set_weights(new_weights)
            i = i + 2

        else:
            x = layer(x)
            i = i + 1

    if not isinstance(model.layers[-1], BatchNormalization):
        x = model.layers[-1](x)

    return Model(inputs=model.input, outputs=[x], name='model_foldBN')


def merge_separable_conv(model):
    """Returns a new model where all depthwise conv2d layers followed by conv2d
    layers are merged into single separable conv layers.

    The new model is strictly equivalent to the previous one.

    Args:
        model (:obj:`tf.keras.Model`): a Keras model.

    Returns:
        :obj:`tf.keras.Model`: a tf.keras.Model.

    """
    # If no layers are Depthwise, there is nothing to be done, return.
    if not any([isinstance(l, QuantizedDepthwiseConv2D) for l in model.layers]):
        return model

    if isinstance(model.layers[0], InputLayer):
        x = model.layers[0].output
        i = 1
    else:
        x = model.layers[0].input
        i = 0
    while i < len(model.layers) - 1:
        layer = model.layers[i]
        next_layer = model.layers[i+1]

        if isinstance(layer, QuantizedDepthwiseConv2D):
            # Check layers expected order
            if not isinstance(next_layer, QuantizedConv2D):
                raise AttributeError(f"Layer {layer.name} "
                                     "QuantizedDepthwiseConv2D should be "
                                     "followed by QuantizedConv2D layers.")

            if layer.bias is not None:
                raise AttributeError(f"Unsupported bias in "
                                     "QuantizedDepthwiseConv2D Layer "
                                     "{layer.name} ")

            # Get weights and prepare new ones
            dw_weights = layer.get_weights()[0]
            pw_weights = next_layer.get_weights()[0]
            new_weights = [dw_weights, pw_weights]
            if next_layer.use_bias:
                bias = next_layer.get_weights()[1]
                new_weights.append(bias)

            # Create new layer
            new_name = f'{layer.name}_{next_layer.name}'
            new_layer = QuantizedSeparableConv2D(next_layer.filters,
                                                 layer.kernel_size,
                                                 quantizer=layer.quantizer,
                                                 padding=layer.padding,
                                                 use_bias=next_layer.use_bias,
                                                 name=new_name)
            x = new_layer(x)
            new_layer.set_weights(new_weights)
            i = i + 2

        else:
            x = layer(x)
            i = i + 1

    # Add last layer if not done already
    if i == (len(model.layers) - 1):
        if isinstance(model.layers[-1], QuantizedDepthwiseConv2D):
            raise AttributeError(f"Layer {layer.name} "
                                 "QuantizedDepthwiseConv2D should be followed "
                                 "by QuantizedConv2D layers.")
        x = model.layers[-1](x)

    return Model(inputs=model.input, outputs=[x], name=model.name)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2019 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

# -*- coding: utf-8 -*-
from __future__ import absolute_import
import tensorflow.keras.backend as K
import tensorflow as tf


class BaseWeightQuantizer:
    """The base class for all Quantizers.

    The relationship between raw weights and quantized weights is:
    (1) W = Wq * scale_factor
    This class provides methods that return quantized weights and their
    associated scale factor.

    """

    def __init__(self, bitwidth):
        """Creates a Weights quantizer for the specified bitwidth.

        Args:
            bitwidth (integer): the quantizer bitwidth defining the number of
                quantized values.

        """
        self.bitwidth_ = bitwidth

    def quantize(self, w):
        """Quantizes the specified weights Tensor.

        Args:
            w (:obj:`tensorflow.Tensor`): the weights Tensor to quantize.

        Returns:
            :obj:`tensorflow.Tensor`: a Tensor of quantized weights.

        """
        return w

    def scale_factor(self, w):
        """Evaluates the scale factor for the specified weights Tensor.

        Args:
          w (:obj:`tensorflow.Tensor`): the weights Tensor to quantize.

        Returns:
          :obj:`tensorflow.Tensor`: a Tensor containing a single scalar value.

        """
        return tf.constant(1.0)

    @property
    def bitwidth(self):
        return self.bitwidth_


class WeightQuantizer(BaseWeightQuantizer):
    """A uniform quantizer.

    Quantize the specified weights into 2^k-1 values, k representing the
    bitwidth.
    The relationship between raw weights and quantized weights is:
    (1) W = Wq * scale_factor

    All values below or above a specified multiple of the weights standard
    deviation are automatically assigned to the min (resp max) value.

    """
    def __init__(self, threshold=3, bitwidth=4):
        """Creates a Weights quantizer for the specified bitwidth.

        Args:
            threshold (integer): the standard deviation multiplier used to exclude
                outliers.
            bitwidth (integer): the quantizer bitwidth defining the number of
                quantized values.

        """
        # Having a cast guarantees a check when the parameters are not numbers
        # (e.g.: None)
        self.threshold_ = float(threshold)
        self.bitwidth_ = int(bitwidth)
        self.kmax_ = (2.**(self.bitwidth_ - 1) - 1)

    def sigma_scaled_(self, w):
        return K.std(w) * self.threshold_

    def bit_range_(self):
        return self.kmax_

    def scale_factor(self, w):
        return self.kmax_ / self.sigma_scaled_(w)

    def quantize(self, w):
        kmax_ = self.kmax_
        delta = self.scale_factor(w)
        out_w = K.clip(round_through(w * delta), -kmax_, kmax_) / delta
        return out_w

class WeightFloat(BaseWeightQuantizer):
    """This quantizer actually does not perform any quantization, and it might
    be used for training.

    """
    def __init__(self, bitwidth = 0):
        super().__init__(bitwidth=bitwidth)


def round_through(x):
    """Element-wise rounding to the closest integer with full gradient propagation.
    A trick from [Sergey Ioffe](http://stackoverflow.com/a/36480182).

    """
    rounded = K.round(x)
    return x + K.stop_gradient(rounded - x)


def ceil_through(x):
    """Element-wise ceiling operation (to the closest greater integer) with
    full gradient propagation.

    """
    ceiling_value = tf.math.ceil(x)
    return x + K.stop_gradient(ceiling_value - x)



# Tensorflow is provided by two packages, tensorflow and tensorflow-gpu.
# This check will check if something is already installed: if that is the case,
# it will just require a given version, otherwise it will throw a RuntimeError
try:
    import tensorflow
    assert int(tensorflow.__version__[0]) >= 2
except:
    raise RuntimeError('cnn2snn requires either tensorflow or tensorflow-gpu '
                       'version >= 2.0 to be installed')

from .utils import invert_batchnorm_pooling, fold_batch_norms, \
    merge_separable_conv
from .model_generator import generate_model
from .converter import convert
from .mapping_generator import check_model_compatibility
from .quantization_ops import WeightQuantizer, WeightFloat
from .quantization_layers import (QuantizedConv2D, QuantizedDepthwiseConv2D,
                                  QuantizedDense, QuantizedSeparableConv2D,
                                  ActivationDiscreteRelu)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2019 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Parsing function to generate layers mapping between Keras and Akida.
Two data classes store the mapping: LayerMapping and ModelMapping.

"""
import tensorflow.keras.layers as layers
from . import quantization_layers as qlayers
from akida import LayerType

class LayerMapping:
    """This data class stores the indices of Keras layers that represent a
    single Akida layer. For example, a 'Convolutional' Akida layer corresponds
    to multiple Keras layers:
    - a QuantizedConv2D layer
    - an optional batch normalization layer
    - an optional pooling layer
    - a discrete ReLU activation (optional if last layer)
    """
    def __init__(self, layer_type, index_neural):
        """Creates a layer map of a single Akida layer from Keras layers.

        Args:
            layer_type (:obj:`akida.LayerType`): the type of the Akida layer.
            index_neural (int): the index of the corresponding Keras
                neural layer.

        """
        self.layer_type = layer_type
        self.index_neural = index_neural
        self.index_pool = None
        self.index_batchnorm = None
        self.index_activation = None


class ModelMapping:
    """This data class maps a Keras model to a future Akida model (not built yet).

    When an instance of ModelMapping is created, it will generate a list of
    LayerMapping objects mapping the Keras model with a succession of Akida
    layers.
    A check is then performed to ensure that the Keras model is compatible with
    Akida.

    Note:
        Note that no Akida model is generated at this step: only a mapping is
        created.

    """
    def __init__(self, model_keras, layer_maps):
        self.model_keras = model_keras
        self.layer_maps = layer_maps


def generate_model_mapping(model, input_is_sparse):
    """Generates a model map between Keras and Akida models.

    This function returns a model map from a Keras model. The model map
    corresponds to the Akida layers mapped from the Keras layers.

    Args:
        model (tf.keras model): the model to parse.
        input_is_sparse (bool): if True, input will be an inputData layer,
            otherwise it will be inputConvolutional.

    Returns:
       :obj:`ModelMapping`: a model map corresponding to the input Keras model.

    """
    for layer in model.layers:
        if hasattr(layer, 'data_format'):
            if layer.data_format == "channels_first":
                raise RuntimeError("unsupported data format channels_first")

    layer_maps = []
    index_unsupported_activations = []

    # First we need to map the input layer
    first_layer = 0
    # If first layer is input layer, skip it
    if isinstance(model.layers[0], layers.InputLayer):
        first_layer = 1
    # Get first Akida layer
    if input_is_sparse:
        layer_ak = LayerMapping(LayerType.InputData, first_layer)
    else:
        layer_ak = LayerMapping(LayerType.InputConvolutional, first_layer)
        first_layer += 1

    # Loop on layers
    neural_layers = (qlayers.QuantizedConv2D, qlayers.QuantizedSeparableConv2D,
                     qlayers.QuantizedDense)
    ignore_list = (layers.Dropout)
    activation_list = (layers.Activation, layers.Softmax)
    for i in range(first_layer, len(model.layers)):
        layer = model.layers[i]
        # If this layer is a neural layer, append the current Akida layer and
        # start a new one
        if isinstance(layer, neural_layers):
            layer_maps.append(layer_ak)
            layer_neural = layer

        # Neural layers
        if isinstance(layer, qlayers.QuantizedConv2D):
            layer_ak = LayerMapping(LayerType.Convolutional, i)
        elif isinstance(layer, qlayers.QuantizedSeparableConv2D):
            layer_ak = LayerMapping(LayerType.SeparableConvolutional, i)
        elif isinstance(layer, qlayers.QuantizedDense):
            layer_ak = LayerMapping(LayerType.FullyConnected, i)

        # Pooling + batchnorm + activation layers
        elif isinstance(layer, (layers.MaxPooling2D, layers.GlobalAveragePooling2D)):
            if layer_ak.index_pool:
                raise RuntimeError(f"Two pooling layers were detected in layer"
                                   f" {layer_neural.name}. Only one pooling "
                                   f"layer is supported.")
            layer_ak.index_pool = i
        elif isinstance(layer, layers.BatchNormalization):
            if layer_ak.index_batchnorm:
                raise RuntimeError(f"Two BatchNormalization layers were "
                                   f"detected after layer {layer_neural.name}."
                                   f" Only one BatchNormalization layer is "
                                   f"supported.")
            layer_ak.index_batchnorm = i
        elif isinstance(layer, qlayers.BaseQuantizedActivation):
            if layer_ak.index_activation:
                raise RuntimeError(f"Two quantized activation layers were "
                                   f"detected after layer {layer_neural.name}."
                                   f" Only one activation layer is supported.")
            layer_ak.index_activation = i

        # Allow flatten before a dense layer
        elif isinstance(layer, layers.Flatten):
            try:
                if isinstance(model.layers[i+1], qlayers.QuantizedDense):
                    continue
            except IndexError:
                pass
            raise RuntimeError("Flatten layer only supported before a Dense "
                               "one")
        # Check Reshape compatibility
        elif isinstance(layer, layers.Reshape):
            _check_reshape_layer(layer)
        # Get unsupported activation index to check compatibility later
        elif isinstance(layer, activation_list):
            index_unsupported_activations.append(i)
        # Allow some other layers useful in keras but that will be discarded
        # or ignored during conversion
        elif isinstance(layer, ignore_list):
            continue
        else:
            # If you got here it means the layer is not recognised: raise an error.
            raise RuntimeError(f"layer {layer.name}: unsupported type "
                               f"{layer.__class__.__name__}.")

    # Append last parsed layer if any
    layer_maps.append(layer_ak)

    # Check if the unsupported activation is after the last neural layer (i.e.
    # in the last Akida layer
    for i in index_unsupported_activations:
        index_last = layer_maps[-1].index_neural
        if i < index_last:
            raise RuntimeError("Activation layers other than quantized "
                               "activations are not supported before the last "
                               "neural layer. Receives activation layer "
                               f"'{model.layers[i].name}' before the last neural"
                               f" layer '{model.layers[index_last].name}'")

    return ModelMapping(model, layer_maps)


def check_mapping_compatibility(model_map):
    """Checks whether the future model will be compatible with Akida.

    This function must mainly test the incompatibities due to the Keras layers
    and the order of the layers (parameters of the quantized layers have
    already been tested at their creation).

    Args:
        :obj:`ModelMapping`: a model map corresponding to the Keras model to
            check
    """
    layers_k = model_map.model_keras.layers

    # Error if hidden layer without activation
    for layer_map in model_map.layer_maps[:-1]:
        if layer_map.layer_type != LayerType.InputData \
                and not layer_map.index_activation:
            raise RuntimeError("No activation layer detected with layer "
                               f"{layers_k[layer_map.index_neural].name}. "
                               "Activation is required in hidden layers.")

    for layer_map in model_map.layer_maps:
        layer_neural = layers_k[layer_map.index_neural]

        # Raise error if BatchNormalization 'axis' is different from the last
        # dimension. The 'axis' parameter is a list containing the axes on what
        # the batch normalization is applied.
        if layer_map.index_batchnorm:
            layer_BN = layers_k[layer_map.index_batchnorm]
            axis = layer_BN.axis
            if (len(layer_BN.axis) != 1 or
                    layer_BN.axis[0] != len(layer_BN.input_shape) - 1):
                raise RuntimeError(f"The BatchNormalization layer "
                                   f"{layer_BN.name} must be applied on the "
                                   f"last axis. Receives {layer_BN.axis}.")

        # Raise error if BatchNormalization is placed after the activation
        if (layer_map.index_batchnorm and layer_map.index_activation and
                layer_map.index_batchnorm > layer_map.index_activation):
            raise RuntimeError(f"In the layer {layer_neural.name}, the batch "
                               "normalization layer must be placed before "
                               "the activation.")

        # Raise error if GlobalAvgPool2D is placed after the activation
        if (layer_map.index_pool and layer_map.index_activation and
                isinstance(layers_k[layer_map.index_pool], layers.GlobalAvgPool2D)
                and layer_map.index_pool > layer_map.index_activation):
            raise RuntimeError(f"In the layer {layer_neural.name}, the global "
                               "average pooling layer must be placed before "
                               "the activation.")

        # Raises error if QuantizedDense input shape is incorrect: supported
        # shapes are (N,) and (1, 1, N). Remember input_shape has the batch
        # size as first element of tuple.
        if isinstance(layer_neural, qlayers.QuantizedDense):
            valid = ( # Input shape is (N,)
                    len(layer_neural.input_shape) == 2
                    or
                    # Input shape is (1, 1, N)
                    (len(layer_neural.input_shape) == 4 and
                    layer_neural.input_shape[1] == 1 and
                    layer_neural.input_shape[2] == 1
                    ))
            if not valid:
                raise RuntimeError("The QuantizedDense layer "
                                   f"{layer_neural.name} must have an input "
                                   "shape of (N,). Receives "
                                   f"{layer_neural.input_shape[1:]}.")

        # Raises error if the padding of MaxPool2D is different from the padding
        # of the neural processing layer.
        if layer_map.index_pool:
            layer_pool = layers_k[layer_map.index_pool]
            if (isinstance(layer_pool, layers.MaxPool2D) and
                    layer_neural.padding != layer_pool.padding):
                raise RuntimeError(f"Pooling layer {layer_pool.name} (padding: "
                                   f"{layer_pool.padding}) must have the same "
                                   f"padding as {layer_neural.name} (padding: "
                                   f"{layer_neural.padding}).")


def _check_reshape_layer(layer):
    """This function checks if the reshape layer is supported.

    In the cnn2snn conversion, a Reshape layer can only be used to transform
    a tensor of shape (N,) to a tensor of shape (1, 1, N), and vice-versa.

    Note that the 'input_shape' and 'output_shape' parameters of a layer has
    the batch size as first element:
        input_shape = (batch_size,) + input_tensor_shape
    The batch size is ignored in the following function.
    """
    in_shape = layer.input_shape
    out_shape = layer.output_shape

    valid = ((# Reshape from (1,1,N) to (N,)
            len(in_shape) == 4 and
            in_shape[1] == 1 and
            in_shape[2] == 1 and
            len(out_shape) == 2 and
            out_shape[1] == in_shape[3])
            or
            # Reshape from (N,) to (1,1,N)
            (len(in_shape) == 2 and
            len(out_shape) == 4 and
            out_shape[1] == 1 and
            out_shape[2] == 1 and
            out_shape[3] == in_shape[1]))

    if not valid:
        raise RuntimeError(f"The Reshape layer {layer.name} can only be used "
                           "to transform a tensor of shape (N,) to a tensor of "
                           "shape (1, 1, N), and vice-versa. Receives "
                           f"input_shape {in_shape[1:]} and output_shape "
                           f"{out_shape[1:]}.")


def check_model_compatibility(model_keras, input_is_sparse):
    """Checks if a Keras quantized model is compatible for cnn2snn conversion.

    This function doesn't convert the Keras quantized model to an Akida model
    but only checks if the model is compatible. The checks are performed at two
    different levels:

        1. Some checks are done when the Keras model is scanned, during the
           generation of the model map.
        2. Other checks are then done based on the model map.

    **1. How to build a compatible Keras quantized model?**

    The following lines give details and constraints on how to build a Keras
    model compatible for the conversion to an Akida model.


    **2. General information about layers**

    An Akida layer must be seen as a block of Keras layers starting with a
    processing layer (QuantizedConv2D, QuantizedSeparableConv2D,
    QuantizedDense). All blocks of Keras layers except the last block must have
    exactly one quantized activation layer (ActivationDiscreteRelu). Other
    optional layers can be present in a block such as a pooling layer or a
    batch normalization layer.
    Here are all the supported Keras layers for an Akida-compatible model:

    - Processing layers:

      - cnn2snn.QuantizedConv2D
      - cnn2snn.QuantizedSeparableConv2D
      - cnn2snn.QuantizedDense

    - Activation layers:

      - cnn2snn.ActivationDiscreteRelu
      - any increasing activation function (only for the last block of layers)
        such as softmax, sigmoid. This layer must derive from
        tf.keras.layers.Activation.

    - Pooling layers:

      - MaxPool2D
      - GlobalAvgPool2D

    - BatchNormalization
    - Dropout
    - Flatten
    - Input
    - Reshape

    Example of a block of Keras layers::

          -------------------
          | QuantizedConv2D |
          -------------------
                  ||
                  \/
        ----------------------
        | BatchNormalization |
        ----------------------
                  ||
                  \/
             -------------
             | MaxPool2D |
             -------------
                  ||
                  \/
       --------------------------
       | ActivationDiscreteRelu |
       --------------------------


    **3. Constraints about inputs**

    An Akida model can accept two types of inputs: sparse events or 8-bit
    images. Whatever the input type, the Keras inputs must respect the
    following relation:

        input_akida = scale * input_keras + shift

    where the Akida inputs must be positive integers, the input scale must be
    a float value and the input shift must be an integer. In other words,
    scale * input_keras must be integers.

    Depending on the input type:

    - if the inputs are events (sparse), the first layer of the Keras model can
      be any quantized processing layer. The input shift must be zero.
    - if the inputs are images, the first layer must be a QuantizedConv2D
      layer.


    **4. Constraints about layers' parameters**

    To be Akida-compatible, the Keras layers must observe the following rules:

    - all layers with the 'data_format' parameter must be 'channels_last'
    - all processing quantized layers and ActivationDiscreteRelu must have a
      valid quantization bitwidth
    - a QuantizedConv2D layer must have strides of (1, 1). Exception: if the
      inputs are images, the first convolutional layer can have any value
      for strides.
    - a QuantizedDense layer must have an input shape of (N,) or (1, 1, N)
    - a BatchNormalization layer must have 'axis' set to -1 (default)
    - a BatchNormalization layer cannot have negative gammas
    - Reshape layers can only be used to transform a tensor of shape (N,) to a
      tensor of shape (1, 1, N), and vice-versa
    - only one pooling layer can be used in each block
    - a MaxPool2D layer must have the same 'padding' as the corresponding
      processing quantized layer

    **5. Constraints about the order of layers**

    To be Akida-compatible, the order of Keras layers must observe the following
    rules:

    - a block of Keras layers must start with a processing quantized layer
    - where present, a BatchNormalization/GlobalAvgPool2D layer must be placed
      before the activation
    - a Flatten layer can only be used before a QuantizedDense layer
    - an Activation layer other than ActivationDiscreteRelu can only be used
      in the last layer


    Args:
        model (tf.keras model): the model to parse.
        input_is_sparse (bool): if True, input will be an inputData layer,
            otherwise it will be inputConvolutional.
    """
    try:
        model_map = generate_model_mapping(model_keras, input_is_sparse)
        check_mapping_compatibility(model_map)
        return True
    except Exception as e:
        print("The Keras quantized model is not compatible for a conversion "
              "to an Akida model:\n", str(e))
        return False



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2019 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

# Tensorflow imports
from tensorflow.python.keras import backend
import tensorflow.keras.backend as K
from tensorflow.keras import layers
from tensorflow.python.keras.utils import conv_utils
from tensorflow.python.framework import ops, common_shapes
from tensorflow.python.ops import nn
from tensorflow.python.ops import standard_ops, gen_math_ops
from tensorflow.python.eager import context
from .quantization_ops import WeightFloat, BaseWeightQuantizer, ceil_through


def _check_unsupported_args(kwargs, unsupported_args):
    """Raises error if unsupported argument are present in kwargs.

    For now, 4 arguments are unsupported: 'data_format', 'activation',
        'dilation_rate', 'depth_mutiplier', 'activity_regularizer'.

    Args:
        kwargs (dictionary): keyword arguments to check.
        unsupported_args: list of unsupported arguments.

    """
    for kwarg in kwargs:
        if kwarg in unsupported_args:
            raise TypeError(
                "Unsupported argument in quantized layers:", kwarg)


class QuantizedConv2D(layers.Conv2D):
    """A quantization-aware Keras convolutional layer.

    Inherits from Keras Conv2D layer, applying a quantization on weights during
    the forward pass.

    """
    def __init__(self,
                 filters,
                 kernel_size,
                 strides=(1, 1),
                 padding='valid',
                 use_bias=True,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 quantizer=WeightFloat(),
                 **kwargs):
        """Creates a quantization-aware convolutional layer.

        Args:
            filters (integer): the number of filters.
            kernel_size (tuple of integer): the kernel spatial dimensions.
            strides (integer, or tuple of integers, optional): strides of the
                convolution along height and width.
            padding (str, optional): one of 'valid' or 'same'.
            use_bias (boolean, optional): whether the layer uses a bias vector.
            kernel_initializer (str, or a :obj:`tf.keras.initializer`, optional):
                initializer for the weights matrix.
            bias_initializer (str, or a :obj:`tf.keras.initializer`, optional):
                initializer for the bias vector.
            kernel_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
                regularization applied to the weights.
            bias_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
                regularization applied to the bias.
            kernel_constraint (str, or a :obj:`tf.keras.constraint`, optional):
                constraint applied to the weights.
            bias_constraint (str, or a :obj:`tf.keras.constraint`, optional):
                constraint applied to the bias.
            quantizer (:obj:`cnn2snn.WeightQuantizer`, optional): the quantizer
                to apply during the forward pass.

        """
        unsupported_args = {
            'data_format',
            'activation',
            'dilation_rate',
            'activity_regularizer'
        }
        _check_unsupported_args(kwargs, unsupported_args)
        if not isinstance(quantizer, BaseWeightQuantizer):
            raise AttributeError("Quantizer object should be used")
        self.quantizer = quantizer
        super(QuantizedConv2D, self).__init__(
                filters=filters,
                kernel_size=kernel_size,
                strides=strides,
                padding=padding,
                use_bias=use_bias,
                kernel_initializer=kernel_initializer,
                bias_initializer=bias_initializer,
                kernel_regularizer=kernel_regularizer,
                bias_regularizer=bias_regularizer,
                kernel_constraint=kernel_constraint,
                bias_constraint=bias_constraint,
                **kwargs)

    def call(self, inputs):
        """Evaluates input Tensor.

        This applies the quantization on weights, then evaluates the input
        Tensor and produces the output Tensor.

        Args:
            inputs(:obj:`tensorflow.Tensor`): input Tensor.

        Returns:
            :obj:`tensorflow.Tensor`: output Tensor.

        """
        outputs = self._convolution_op(inputs,
                                       self.quantizer.quantize(self.kernel))
        if self.use_bias:
            if self.data_format == "channels_first":
                raise RuntimeError("unsupported data format channels_first")
            else:
                outputs = nn.bias_add(outputs, self.bias, data_format='NHWC')

        return outputs


class QuantizedDepthwiseConv2D(layers.DepthwiseConv2D):
    """A quantization-aware Keras depthwise convolutional layer.

    Inherits from Keras DepthwiseConv2D layer, applying a quantization on
    weights during the forward pass.

    """
    def __init__(self,
                 kernel_size,
                 padding='valid',
                 use_bias=True,
                 depthwise_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 depthwise_regularizer=None,
                 bias_regularizer=None,
                 depthwise_constraint=None,
                 bias_constraint=None,
                 quantizer=WeightFloat(),
                 **kwargs):
        """Creates a quantization-aware depthwise convolutional layer.

        Args:
            kernel_size (a tuple of integer): the kernel spatial dimensions.
            padding (str, optional): One of 'valid' or 'same'.
            use_bias (boolean, optional): whether the layer uses a bias vector.
            depthwise_initializer (str, or a :obj:`tf.keras.initializer`, optional):
                initializer for the weights matrix.
            bias_initializer (str, or a :obj:`tf.keras.initializer`, optional):
                initializer for the bias vector.
            depthwise_regularizer (str, or a :obj:`tf.keras.initializer`, optional):
                regularization applied to the weights.
            bias_regularizer (str, or a :obj:`tf.keras.initializer`, optional):
                regularization applied to the bias.
            depthwise_constraint (str, or a :obj:`tf.keras.initializer`, optional):
                constraint applied to the weights.
            bias_constraint (str, or a :obj:`tf.keras.initializer`, optional):
                constraint applied to the bias.
            quantizer (:obj:`cnn2snn.WeightQuantizer`, optional): the quantizer
                to apply during the forward pass.

        """
        unsupported_args = {
            'data_format',
            'activation',
            'strides',
            'depth_multiplier',
            'activity_regularizer'
        }
        _check_unsupported_args(kwargs, unsupported_args)
        if not isinstance(quantizer, BaseWeightQuantizer):
            raise AttributeError("Quantizer object should be used")
        self.quantizer = quantizer
        super(QuantizedDepthwiseConv2D, self).__init__(
                kernel_size=kernel_size,
                padding=padding,
                use_bias=use_bias,
                depthwise_initializer=depthwise_initializer,
                bias_initializer=bias_initializer,
                depthwise_regularizer=depthwise_regularizer,
                bias_regularizer=bias_regularizer,
                depthwise_constraint=depthwise_constraint,
                bias_constraint=bias_constraint,
                **kwargs)

    def call(self, inputs):
        """Evaluates input Tensor.

        This applies the quantization on weights, then evaluates the input
        Tensor and produces the output Tensor.

        Args:
            inputs (:obj:`tensorflow.Tensor`): input Tensor.

        Returns:
            :obj:`tensorflow.Tensor`: output Tensor.

        """
        # We don't support biases
        return backend.depthwise_conv2d(inputs,
                                        self.quantizer.quantize(
                                            self.depthwise_kernel),
                                        strides=self.strides,
                                        padding=self.padding,
                                        dilation_rate=self.dilation_rate,
                                        data_format=self.data_format)


class QuantizedDense(layers.Dense):
    """A quantization-aware Keras dense layer.

    Inherits from Keras Dense layer, applying a quantization on weights during
    the forward pass.

    """
    def __init__(self,
                 units,
                 use_bias=True,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 quantizer=WeightFloat(),
                 **kwargs):
        """Creates a quantization-aware dense layer.

        Args:
            units (integer): the number of neurons.
            use_bias (boolean, optional): whether the layer uses a bias vector.
            kernel_initializer (str, or a :obj:`tf.keras.initializer`, optional):
                initializer for the weights matrix.
            bias_initializer (str, or a :obj:`tf.keras.initializer`, optional):
                initializer for the bias vector.
            kernel_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
                regularization applied to the weights.
            bias_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
                regularization applied to the bias.
            kernel_constraint (str, or a :obj:`tf.keras.constraint`, optional):
                constraint applied to the weights.
            bias_constraint (str, or a :obj:`tf.keras.constraint`, optional):
                constraint applied to the bias.
            quantizer (:obj:`cnn2snn.WeightQuantizer`, optional): the quantizer
                to apply during the forward pass.

        """
        unsupported_args = {
            'activation',
            'activity_regularizer'
        }
        _check_unsupported_args(kwargs, unsupported_args)
        if not isinstance(quantizer, BaseWeightQuantizer):
            raise AttributeError("Quantizer object should be used")
        self.quantizer = quantizer
        super(QuantizedDense, self).__init__(
                units=units,
                use_bias=use_bias,
                kernel_initializer=kernel_initializer,
                bias_initializer=bias_initializer,
                kernel_regularizer=kernel_regularizer,
                bias_regularizer=bias_regularizer,
                kernel_constraint=kernel_constraint,
                bias_constraint=bias_constraint,
                **kwargs)

    def call(self, inputs):
        """Evaluates input Tensor.

        This applies the quantization on weights, then evaluates the input
        Tensor and produces the output Tensor.

        Args:
            inputs (:obj:`tensorflow.Tensor`): input Tensor.

        Returns:
            :obj:`tensorflow.Tensor`: output Tensor.

        """
        inputs = ops.convert_to_tensor(inputs)
        kernel = self.quantizer.quantize(self.kernel)
        rank = common_shapes.rank(inputs)
        if rank > 2:
            # Broadcasting is required for the inputs.
            outputs = standard_ops.tensordot(inputs,
                                             kernel, [[rank - 1], [0]])
            # Reshape the output back to the original ndim of the input.
            if not context.executing_eagerly():
                shape = inputs.get_shape().as_list()
                output_shape = shape[:-1] + [self.units]
                outputs.set_shape(output_shape)
        else:
            outputs = gen_math_ops.mat_mul(inputs, kernel)
        if self.use_bias:
            outputs = nn.bias_add(outputs, self.bias)
        return outputs


class QuantizedSeparableConv2D(layers.SeparableConv2D):
    """A quantization-aware Keras separable convolutional layer.

    Inherits from Keras SeparableConv2D layer, applying a quantization on
    weights during the forward pass.

    """
    def __init__(self,
                 filters,
                 kernel_size,
                 padding='valid',
                 use_bias=True,
                 depthwise_initializer='glorot_uniform',
                 pointwise_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 depthwise_regularizer=None,
                 pointwise_regularizer=None,
                 bias_regularizer=None,
                 depthwise_constraint=None,
                 pointwise_constraint=None,
                 bias_constraint=None,
                 quantizer=WeightFloat(),
                 **kwargs):
        """Creates a quantization-aware separable convolutional layer.

        Args:
            filters (integer): the number of filters.
            kernel_size (tuple of integer): the kernel spatial dimensions.
            padding (str, optional): One of 'valid' or 'same'.
            use_bias (boolean, optional): Whether the layer uses a bias vector.
            depthwise_initializer (str, or a :obj:`tf.keras.initializer`, optional):
                initializer for the depthwise kernel.
            pointwise_initializer (str, or a :obj:`tf.keras.initializer`, optional):
                initializer for the pointwise kernel.
            bias_initializer (str, or a :obj:`tf.keras.initializer`, optional):
                initializer for the bias vector.
            depthwise_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
                regularization applied to the depthwise kernel.
            pointwise_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
                regularization applied to the pointwise kernel.
            bias_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
                regularization applied to the bias.
            depthwise_constraint (str, or a :obj:`tf.keras.constraint`, optional):
                constraint applied to the depthwise kernel.
            pointwise_constraint (str, or a :obj:`tf.keras.constraint`, optional):
                constraint applied to the pointwise kernel.
            bias_constraint (str, or a :obj:`tf.keras.constraint`, optional):
                constraint applied to the bias.
            quantizer (:obj:`cnn2snn.WeightQuantizer`): the quantizer to apply
                during the forward pass.

        """
        unsupported_args = {
            'data_format',
            'activation',
            'strides',
            'dilation_rate',
            'depth_multiplier',
            'activity_regularizer'
        }
        _check_unsupported_args(kwargs, unsupported_args)
        if not isinstance(quantizer, BaseWeightQuantizer):
            raise AttributeError("Quantizer object should be used")
        self.quantizer = quantizer
        super(QuantizedSeparableConv2D, self).__init__(
                filters=filters,
                kernel_size=kernel_size,
                padding=padding,
                use_bias=use_bias,
                depthwise_initializer=depthwise_initializer,
                pointwise_initializer=pointwise_initializer,
                bias_initializer=bias_initializer,
                depthwise_regularizer=depthwise_regularizer,
                pointwise_regularizer=pointwise_regularizer,
                bias_regularizer=bias_regularizer,
                depthwise_constraint=depthwise_constraint,
                pointwise_constraint=pointwise_constraint,
                bias_constraint=bias_constraint,
                **kwargs)

    def call(self, inputs):
        """Evaluates input Tensor.

        This applies the quantization on weights, then evaluates the input
        Tensor and produces the output Tensor.

        Args:
            inputs (:obj:`tensorflow.Tensor`): input Tensor.

        Returns:
            :obj:`tensorflow.Tensor`: a Tensor.

        """
        strides = (1,) + self.strides + (1,)
        outputs = nn.separable_conv2d(
            inputs,
            self.quantizer.quantize(
                self.depthwise_kernel),
            self.quantizer.quantize(
                self.pointwise_kernel),
            strides=strides,
            padding=self.padding.upper(),
            rate=self.dilation_rate,
            data_format=conv_utils.convert_data_format(self.data_format, ndim=4))

        if self.use_bias:
            outputs = nn.bias_add(
                outputs,
                self.bias,
                data_format=conv_utils.convert_data_format(self.data_format, ndim=4))

        return outputs


class BaseQuantizedActivation(layers.Activation):
    """Base class for quantized activation layers."""

    def __init__(self, activation, **kwargs):
        super().__init__(activation=activation, **kwargs)


class ActivationDiscreteRelu(BaseQuantizedActivation):
    """A discrete ReLU Keras Activation.

    Activations will be quantized and will have 2^bitwidth values in the range
    [0,6].

    """
    def __init__(self, bitwidth=1, **kwargs):
        """Creates a discrete ReLU for the specified bitwidth.

        Args:
            bitwidth (integer): the activation bitwidth.

        """
        levels = 2.**bitwidth - 1
        self.relumax_ = min(levels, 6)
        self.bitwidth = bitwidth
        self.gamma_k = self.relumax_ / levels
        self.t0_k = 0.5 * self.relumax_ / levels
        self.step = (2.**bitwidth * self.gamma_k) / 16
        # Legacy parameters, used by cnn2snn.keras2akida: they might be removed
        # later
        self.scale_factor = levels / self.relumax_
        self.threshold = self.t0_k

        super().__init__(activation=self.activation, **kwargs)

    def activation(self, x):
        """Evaluates the activations for the specified input Tensor.

        Args:
            x (:obj:`tensorflow.Tensor`): the input values.

        """
        ceiled_scaled = ceil_through(x / self.gamma_k - 0.5)
        return K.clip(self.gamma_k * ceiled_scaled, 0, self.relumax_)


# A helper to instantiate a Conv2D layer to which a modifier is assigned
def conv2d(filters, kernel_size, modifier=None, **kwargs):
    layer = QuantizedConv2D(filters, kernel_size, quantizer=modifier, **kwargs)
    return layer

# A helper to instantiate a Dense layer to which a modifier is assigned
def dense(units, modifier=None, **kwargs):
    layer = QuantizedDense(units, quantizer=modifier, **kwargs)
    return layer

# A helper to instantiate a Depthwise Conv2D layer to which a modifier is assigned
def depthwise_conv2d(kernel_size, modifier=None, **kwargs):
    layer = QuantizedDepthwiseConv2D(kernel_size, quantizer=modifier, **kwargs)
    return layer

def separable_conv2d(filters, kernel_size, modifier=None, **kwargs):
    layer = QuantizedSeparableConv2D(filters, kernel_size, quantizer=modifier, **kwargs)
    return layer

def batchNormalization(*args, **kwargs):
    return layers.BatchNormalization(*args, **kwargs)


def maxPooling2D(*args, **kwargs):
    return layers.MaxPooling2D(*args, **kwargs)


def activationFloat(type='relu'):
    return lambda **kwargs: layers.Activation(type, **kwargs)


def activationDiscreteRelu(bitwidth=1, **kwargs):
    return ActivationDiscreteRelu(bitwidth, **kwargs)

def activationDiscreteReluBits(bitwidth=1):
    return lambda **kwargs: activationDiscreteRelu(bitwidth, **kwargs)

def batchNormalization(*args, **kwargs):
    return layers.BatchNormalization(*args, **kwargs)



from os import path
from setuptools import setup

# Read the contents of the README file
directory = path.abspath(path.dirname(__file__))
with open(path.join(directory, 'README'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='cnn2snn',
      version='1.7.2',
      description='Keras to Akida CNN Converter',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='Alvaro Moran',
      author_email='amoran@brainchip.com',
      url='https://www.akidalab.com',
      license='Apache 2.0',
      packages=['cnn2snn'],
      install_requires=['akida==1.7.2', 'numpy'],
     )



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2019 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert weights from Keras to Akida.

"""
import numpy as np
import tensorflow as tf
import tensorflow.keras.layers as layers
from cnn2snn import quantization_layers as qlayers
from akida import LayerType


def set_weights_thresholds(model_map, model_ak, input_scaling):
    """Converts and sets weights, fire thresholds and threshold steps to
    an Akida model from a model map.

    Args:
        model_map (ModelMapping): a model map that represents the Akida layers
            mapped to a Keras model
        model_ak (:obj:`akida.Model`): an Akida model that has been generated from the
            corresponding model map
        input_scaling (2-element tuple): the input factor and shift

    """
    layers_keras = model_map.model_keras.layers
    input_factor, input_shift = input_scaling

    # Loop over the Akida layers
    current_input_scaling = input_scaling
    for i in range(len(model_map.layer_maps)):
        layer_map = model_map.layer_maps[i]
        layer_ak = model_ak.get_layer(i)

        if layer_map.layer_type == LayerType.InputData:
            continue

        # Convert and set weights
        layer_k = layers_keras[layer_map.index_neural]
        weights, weights_scale_factor = \
            convert_layer_weights(layer_map.layer_type, layer_k,
                                  layer_ak.input_dims)

        if layer_map.layer_type == LayerType.SeparableConvolutional:
            layer_ak.set_variable('weights', weights[0])
            layer_ak.set_variable('weights_pw', weights[1])
        else:
            layer_ak.set_variable('weights', weights)

        # Convert and set fire thresholds and steps
        th_fire, th_step = convert_layer_thresholds(layer_map,
                                                    layers_keras,
                                                    current_input_scaling,
                                                    weights_scale_factor,
                                                    weights)
        layer_ak.set_variable('threshold_fire', th_fire)
        layer_ak.set_variable("threshold_fire_step", th_step)
        if layer_map.index_activation:
            current_input_scaling = (1 / \
                layers_keras[layer_map.index_activation].gamma_k, 0)


def convert_layer_weights(layer_type, layer_k, input_dims=None):
    """Converts weights for an Akida layer.

    This function returns the converted weights for any Akida layer,
    according to the Akida layer type and the Keras neural layer.

    Args:
        layer_type (str): type of the layer Akida.
        layer_k (:obj:`tf.keras.Layer`): the Keras layer from which the weights
            will be converted.
        input_dims (tuple): input dimensions of the Keras layer. This argument
            is only required for a fullyConnected Akida layer.

    Returns:
        tuple: the converted weights and the weights scale factor from
            quantization.

    """
    if layer_type == LayerType.InputConvolutional:
        return _convert_conv_weights(layer_k, input_conv=True)
    elif layer_type == LayerType.Convolutional:
        return _convert_conv_weights(layer_k)
    elif layer_type == LayerType.SeparableConvolutional:
        return _convert_separable_conv_weights(layer_k)
    elif layer_type == LayerType.FullyConnected:
        return _convert_dense_weights(layer_k, input_dims)


def _convert_conv_weights(layer, input_conv=False):
    # Quantize weights
    weights = layer.get_weights()[0]
    wq, delta = quantize_weights(layer.quantizer, weights)

    # Transpose weights to get from Keras HWCN to Akida WHCN
    # and multiply by the weights scale factor.
    wq_akida = wq.transpose((1, 0, 2, 3)) * delta
    wq_akida = np.round(wq_akida).astype(np.int8)

    # Flip W and H dimensions for conv. kernels (not input conv.)
    if not input_conv:
        wq_akida = np.flip(wq_akida, axis=[0, 1])

    return wq_akida, delta


def _convert_separable_conv_weights(layer):
    # Quantize depthwise weights
    w_ak_dw, delta_dw = _convert_conv_weights(layer)

    # Quantize pointwise weights
    weights_pw = layer.get_weights()[1]
    wq_pw, delta_pw = quantize_weights(layer.quantizer, weights_pw)
    # Pointwise weights in Keras have HWCN format and H = W = 1. This
    # makes the conversion to Akida's NCHW trivial.
    w_ak_pw = wq_pw * delta_pw
    w_ak_pw = np.round(w_ak_pw).astype(np.int8)

    # For separable, the weights scale factor can be seen as the product of the
    # DW and PW scale factors.
    delta = delta_dw * delta_pw
    return (w_ak_dw, w_ak_pw), delta


def _convert_dense_weights(layer, input_dims):
    # Quantize weights
    weights = layer.get_weights()[0]
    wq, delta = quantize_weights(layer.quantizer, weights)

    # retrieve input dimensions from Akida's layer
    inwidth, inheight, inchans = input_dims
    # Kernels in the fully connected are in the (HxWxC,N) format, more
    # specifically in each neuron data is laid out in the H,W,C format. In
    # Akida we expect a kernel in the (N,CxHxW,1,1), where the data in each
    # neuron is laid out in the W,H,C format as the input dimensions are
    # set.
    # So the operations done in order to obtain the akida fully connected
    # kernel are:
    # 1. reshape to H,W,C,N to split data across dimensions
    # 2. transpose dimensions to obtain C,H,W,N
    # 3. reshape to: 1,1,CxHxW,N
    #
    wq_akida = wq.reshape(inheight, inwidth, inchans, layer.units) \
        .transpose(2, 0, 1, 3) \
        .reshape(1, 1, inchans * inheight * inwidth, layer.units)
    # Multiply by delta, round and cast to int
    wq_akida = wq_akida * delta
    wq_akida = np.round(wq_akida).astype(np.int8)

    return wq_akida, delta


def convert_layer_thresholds(layer_map,
                             layers_keras,
                             input_scaling,
                             weights_scale_factor,
                             weights_ak):
    """Returns the fire thresholds and threshold steps for a given Akida layer.

    This function computes the fire thresholds and steps for a given Akida layer
    (defined by the layer map) from the corresponding Keras layers. Other
    arguments are required to compute these variables.

    Args:
        layer_map (LayerMapping): the given Akida layer map
        layers_keras: the list of Keras layers (tf.keras.model.layers)
            corresponding to the given layer map
        input_scaling (2-element tuple): the input scaling of the current layer.
        weights_scale_factor (float): the weights scale factor of the current
            Akida layer. This is given by the 'convert_layer_weights' function.
        weights_ak (np.ndarray): the Akida weights of the current Akida layer.

    Returns:
        tuple: the fire thresholds and steps.

    """
    input_scale, input_shift = input_scaling
    layer_neural = layers_keras[layer_map.index_neural]
    threshold_shape = layer_neural.output_shape[-1]
    bias = np.zeros(threshold_shape)
    thresh_keras = np.zeros(threshold_shape)
    step_keras = np.ones(threshold_shape)
    gamma_bn = np.ones(threshold_shape)
    sigma_bn = np.ones(threshold_shape)
    beta_bn = np.zeros(threshold_shape)
    mu_bn = np.zeros(threshold_shape)

    # Get bias if present
    if layer_neural.use_bias:
        bias = layer_neural.get_weights()[-1]

    # Get parameters from the discrete ReLU activation (t0 and step)
    if layer_map.index_activation:
        act_layer = layers_keras[layer_map.index_activation]
        thresh_keras = act_layer.t0_k * np.ones(threshold_shape)
        step_keras = act_layer.step * np.ones(threshold_shape)

    # Get batchnorm parameters
    if layer_map.index_batchnorm:
        layer_bn = layers_keras[layer_map.index_batchnorm]
        (gamma_bn, beta_bn, mu_bn, var_bn) = layer_bn.get_weights()
        eps = layer_bn.epsilon
        sigma_bn = np.sqrt(var_bn + eps)

    # Compute threshold fires and steps
    th_fire = input_scale * weights_scale_factor * \
        (sigma_bn / gamma_bn * (thresh_keras - beta_bn) + mu_bn - bias)
    th_step = input_scale * weights_scale_factor * \
        sigma_bn / gamma_bn * step_keras

    # If there is an input_shift at inputConv layer, update th_fire
    if layer_map.layer_type == LayerType.InputConvolutional and input_shift:
        beta_wq_akida = np.sum(weights_ak, axis=(0, 1, 2)) * input_shift
        th_fire += beta_wq_akida

    # If there is a global average pooling, update th fire and step
    if layer_map.index_pool:
        pooling_layer = layers_keras[layer_map.index_pool]
        if isinstance(pooling_layer, layers.GlobalAveragePooling2D):
            shape = pooling_layer.input_shape
            global_avg_pool_factor = int(shape[1]) * int(shape[2])
            th_fire *= global_avg_pool_factor
            th_step *= global_avg_pool_factor

    th_fire = np.floor(th_fire).astype(np.int32)
    th_step = th_step.astype(np.float32)

    return th_fire, th_step


def quantize_weights(quantizer, w):
    """Returns quantized weights and delta as numpy arrays.

    Internally, it uses a tf.function that wraps calls to the quantizer in
    a graph, allowing the weights to be quantized eagerly.

    Args:
        quantizer (:obj:`WeightQuantizer`): the quantizer object.
        w (:obj:`np.ndarray`): the weights to quantize.

    Retruns:
        :obj:`np.ndarray`: the quantized weights `np.ndarray` and the scale
            factor scalar.

    """
    # Helper function to return quantized weights and scale factor
    def quantize(quantizer, w):
        return quantizer.quantize(w), quantizer.scale_factor(w)
    # Wrap quantize helper in a graph to execute it eagerly
    quantize_wrapper = tf.function(quantize)
    # Invoke wrapper with a constant Tensor initialized from the weights
    wq, delta = quantize_wrapper(quantizer, tf.constant(w))
    # Return eager tensors as numpy arrays
    return wq.numpy(), delta.numpy()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2019 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import os
from tensorflow import executing_eagerly
from . import utils
from .mapping_generator import generate_model_mapping, check_mapping_compatibility
from .model_generator import generate_model
from .weights_ops import set_weights_thresholds
from akida import Model


def convert(model, file_path=None, input_scaling=(1.0, 0), input_is_sparse=False):
    """Simple function to convert a Keras model to an Akida one.

    These steps are performed:

    1) Merge the depthwise+conv layers into a separable_conv one.
    2) Generate an Akida model based on that model.
    3) Convert weights from the Keras model to Akida.

    Note:
        The relationship between Keras and Akida inputs is:
        input_akida = alpha * input_keras + beta, optional

    Args:
        model (:obj:`tf.keras.Model`): a tf.keras model
        file_path (str, optional): destination for the akida model.
            (Default value = None)
        input_scaling (2 elements tuple, optional): value of the input scaling.
            (Default value = (1.0,0))
        input_is_sparse (bool, optional): if True, input will be an InputData layer,
            otherwise it will be InputConvolutional. (Default value = False)

    Returns:
        an Akida model.

    """

    if not executing_eagerly():
        raise SystemError("Tensorflow eager execution is disabled. "
                          "It is required to convert Keras weights to Akida.")

    # Merge separable convolution
    model_sep = utils.merge_separable_conv(model)

    # Generate model mapping
    model_map = generate_model_mapping(model_sep, input_is_sparse)

    # Check compatibility of the model map
    check_mapping_compatibility(model_map)

    # Generate Akida model
    ak_inst = generate_model(model_map, input_scaling)

    # Convert weights
    set_weights_thresholds(model_map, ak_inst, input_scaling)

    # Save model if file_path is given
    if file_path:
        # Create directories
        dir_name, base_name = os.path.split(file_path)
        if base_name:
            file_root, file_ext = os.path.splitext(base_name)
            if not file_ext:
                file_ext = '.fbz'
        else:
            file_root = model.name
            file_ext = '.fbz'

        if dir_name and not os.path.exists(dir_name):
            os.makedirs(dir_name)

        save_path = os.path.join(dir_name, file_root + file_ext)
        ak_inst.save(save_path)

    return ak_inst
